#include <time.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/wait.h>

#define MAX_INPUT_SIZE (1024*1024*11)

#define NOTFALL_REST 10
#define WARN_TIME 10

#define RAUNZ fprintf( stderr, "RAUNZ: code problem at line %d.\n", __LINE__ )

#define RENDER_PATH "./trivial"
#define TREE_PATH "./tree"

time_t start_time, time_limit;

const char *argv0;
const char *notfall_buf = NULL;
ssize_t notfall_buf_len;

struct job {
    struct job *next;
    const char *name;
    pid_t pid;
    int src[2], dst[2];
    int read_fd, write_fd;
    int faulty;
    int running;
    char *buf;
    int maxlen, readlen;
    int pollindex;
};

void install_alarm( int ti )
{
    if (ti < 1)
	ti = 1;
    alarm( ti );
}

struct job *best_job;

static inline int act_time(void)
{
    time_t now;

    time( &now );
    now -= start_time;

    if (now < 0) {
	RAUNZ;
	return 0;
    }
    return (int)now;
}

struct job *spawn( const char *prog_name, int maxlen )
{
    int retval;
    struct job *job = (struct job *)malloc( sizeof( struct job ));

    if (!prog_name) {
	RAUNZ;
	return NULL;
    }

    if (!job) {
	fprintf( stderr, "%s: spawn: malloc(job) failed: %s.\n",
		 argv0, strerror( errno ));
	return NULL;
    }

    job->buf = (char *)malloc( maxlen );
    if (!job->buf) {
	fprintf( stderr, "%s: spawn: malloc(job->buf) failed: %s.\n",
		 argv0, strerror( errno ));
	return NULL;
    }
    job->maxlen = maxlen;
    job->readlen = 0;
    job->next = NULL;
    job->name = prog_name;
    job->faulty = 0;
    retval = pipe( job->src );
    if (retval == -1) {
	fprintf( stderr, "%s: spawn: pipe(src) failed: %s.\n",
		 argv0, strerror( errno ));
	return NULL;
    }
    retval = pipe( job->dst );
    if (retval == -1) {
	fprintf( stderr, "%s: spawn: pipe(dst) failed: %s.\n",
		 argv0, strerror( errno ));
	return NULL;
    }

    job->pid = fork();
    switch (job->pid) {
    case -1:
	fprintf( stderr, "%s: spawn: fork failed: %s\n",
		 argv0, strerror( errno ));
	return NULL;

    case 0: /* child */
	retval = dup2( job->src[0], STDIN_FILENO );
	if (retval == -1) {
	    fprintf( stderr, "%s: spawn/child: dup2(src[0]) failed: %s.\n",
		     argv0, strerror( errno ));
	    _exit( 1 );
	}
	retval = close( job->src[1] );
	if (retval == -1) {
	    fprintf( stderr, "%s: spawn/child: close(src[1]) failed: %s.\n",
		     argv0, strerror( errno ));
	    _exit( 2 );
	}
	retval = dup2( job->dst[1], STDOUT_FILENO );
	if (retval == -1) {
	    fprintf( stderr, "%s: spawn/child: dup2(stdout) failed: %s.\n",
		     argv0, strerror( errno ));
	    _exit( 3 );
	}
	retval = close( job->dst[0] );
	if (retval == -1) {
	    fprintf( stderr, "%s: spawn/child: close(dst[0]) failed: %s.\n",
		     argv0, strerror( errno ));
	    _exit( 4 );
	}
	execlp( prog_name, prog_name, NULL );
	fprintf( stderr, "%s: spawn/child: execlp('%s') failed: %s.\n",
		 argv0, prog_name, strerror( errno ));
	_exit( 5 ); /* failure */

    default: /* parent */
	retval = close( job->src[0] );
	if (retval == -1) {
	    fprintf( stderr, "%s: spawn: spawned job faulty: %s.\n",
		     argv0, strerror( errno ));
	    job->faulty = 1;
	}
	retval = close( job->dst[1] );
	if (retval == -1) {
	    fprintf( stderr, "%s: spawn: spawned job faulty: %s.\n",
		     argv0, strerror( errno ));
	    job->faulty = 1;
	}
	job->read_fd = job->dst[0];
	job->write_fd = job->src[1];
	if ((job->read_fd < 0) || (job->write_fd < 0)) {
	    fprintf( stderr, "%s: spawn: spawned job faults: illegal fds.\n",
		     argv0 );
	    job->faulty = 1;
	}

	if (job->faulty) {
	    fprintf( stderr, "%s: spawn: killing job %d.\n", argv0, job->pid );
	    kill( job->pid, SIGKILL );
	    return NULL;
	}

	fprintf( stderr, "%s: t=%d spawned '%s' as %d.\n",
		 argv0, act_time(), job->name, job->pid );

	job->running = 1;
	return job;
    }
}


int write_try( int fd, const void *buf, size_t count )
{
    int outlen = 0, retval;
    int tries = 5;

    if (count < 0) {
	RAUNZ;
	return -1;
    }
    if (!buf) {
	RAUNZ;
	return -1;
    }

    if (count == 0)
	return 0;

    for (;;) {
	retval = write( fd, buf + outlen, count - outlen );
	if (retval == -1) {
	    tries--;
	    if (tries < 0)
		return -1;
	}
	if (retval > 0) {
	    outlen += retval;
	    if (outlen >= count)
		return count;
	}
    }
}

static inline void kill_job( struct job *job )
{
    if (job) {
	job->running = 0;
	kill( job->pid, SIGKILL );
    } else
	RAUNZ;
}

static inline void warn_job( struct job *job )
{
    if (job)
	kill( job->pid, SIGUSR1 );
    else
	RAUNZ;
}

/* write input data to job and switch read handle to nonblocking */
static int fill_job( struct job *job, const char *buf, int size )
{
    int retval;

    if (!job) {
	RAUNZ;
	return -1;
    }
    if (!buf) {
	RAUNZ;
	return -1;
    }
    if (size < 0) {
	RAUNZ;
	return -1;
    }

    retval = write_try( job->write_fd, buf, size );
    if (retval == -1) {
	fprintf( stderr, "%s: fill_job[%s]: write failed: %s.\n",
		 argv0, job->name, strerror( errno ));
	return -1;
    }
    if (retval != size) {
	fprintf( stderr, "%s: fill_job[%s]: wrote only %d of %d bytes.\n",
		 argv0, job->name, retval, size );
	return -1;
    }
    retval = close( job->write_fd );
    if (retval == -1) {
	fprintf( stderr, "%s: fill_job[%s]: close failed: %s.\n",
		 argv0, job->name, strerror( errno ));
	return -1;
    }

    retval = fcntl( job->read_fd, F_SETFL, O_NONBLOCK );
    if (retval == -1) {
	fprintf( stderr, "%s: fill_job[%s]: fcntl failed: %s.\n",
		 argv0, job->name, strerror( errno ));
	return -1;
    }

    fprintf( stderr, "%s: t=%d filled job '%s'.\n",
		 argv0, act_time(), job->name );
    return 0;
}

char *read_input( ssize_t *len )
{
    char *buf;
    int retval;

    *len = 0;

    buf = (char *)malloc( MAX_INPUT_SIZE + 1 );
    assert( buf );

    for (;;) {
	if ((MAX_INPUT_SIZE - *len) < 0) {
	    fprintf( stderr, "%s: input too big\n", argv0 );
	    exit (66);
	}
	retval = read( STDIN_FILENO, buf + *len, MAX_INPUT_SIZE - *len );
	if (retval == 0)
	    break;
	if (retval > 0)
	    *len += retval;
    }

    buf[*len] = 0;

    return buf;
}

void write_output(void)
{
    int outlen = 0, retval;

    if (notfall_buf_len <= 0)
	return;

    assert( notfall_buf );

    for (;;) {
	retval = write( STDOUT_FILENO, notfall_buf + outlen,
			notfall_buf_len - outlen );
	if (retval == -1)
	    continue;
	if (retval > 0) {
	    outlen += retval;
	    if (outlen >= notfall_buf_len)
		return;
	}
    }
}

struct job *warned_job = NULL;

static int only_warn = 1;

void notfall( int i )
{
    if (only_warn) {
	fprintf( stderr, "%s: t=%d timer hooked for warnings (%s).\n",
		 argv0, act_time(), best_job->name );	only_warn = 0;
	if (warned_job)
	    warn_job( warned_job );
	install_alarm( time_limit - act_time() - NOTFALL_REST );
	return;
    }

    if (best_job && best_job->name)
	fprintf( stderr, "%s: t=%d timer hooked emergency write (%s).\n",
		 argv0, act_time(), best_job->name );
    else
	RAUNZ;

    write_output();
    exit( 0 );
}

void consider_solution( struct job *j )
{
    signal( SIGALRM, SIG_IGN );

    if (j->readlen < notfall_buf_len) { /* switch to better solution */
	best_job = j;
	notfall_buf_len = j->readlen;
	notfall_buf = j->buf;
    }

    signal( SIGALRM, notfall );
    install_alarm( time_limit - act_time() - NOTFALL_REST );
}

void read_jobs( struct job *jobs )
{
    struct pollfd pollfds[10];
    int retval;

    for (;;) {
	struct job *j = jobs;
	int index = 0;

	for (j=jobs; j; j=j->next)
	    if (j->running) {
		pollfds[index].fd = j->read_fd;
		pollfds[index].events = POLLIN;
		pollfds[index].revents = 0;
		j->pollindex = index;
		index++;
	    }

	if (index == 0) {
	    /* no jobs running anymore */
	    return;
	}

	retval = poll( pollfds, index, -1 );
	if (retval == -1) {
	    usleep( 100000 );
	    continue;
	}

	for (j = jobs; j; j = j->next) {
	    if (j->running && (pollfds[j->pollindex].revents & POLLIN)) {
		int retval;

		retval = read( j->read_fd,
			       j->buf + j->readlen, j->maxlen - j->readlen );
		if (retval == -1) {
		    fprintf( stderr, "%s: read_jobs[%s]: read failed: %s.\n",
			     argv0, j->name, strerror( errno ));
		    kill_job( j );
		} else {
		    j->readlen += retval;
		    /*fprintf( stderr, "%s: read_jobs[%s]: got %d bytes.\n",
		      argv0, j->name, retval );*/
		    if (j->readlen >= notfall_buf_len) {
			fprintf( stderr, "%s: read_jobs[%s]: read too much.\n",
				 argv0, j->name);
			kill_job( j ); /* too big */
		    }
		}
	    }
	    if (j->running && (pollfds[j->pollindex].revents & POLLHUP)) {
		int status;
		fprintf( stderr, "%s: read_jobs[%s]: hupped.\n",
			 argv0, j->name );
		waitpid(j->pid, &status, 0);
		if (!(WEXITSTATUS(status) != 0 || WIFSIGNALED(status)))
		    consider_solution( j );
		kill_job( j ); /* sure is sure says the .. */
	    }
	    if (j->running &&
		(pollfds[j->pollindex].revents & (POLLERR|POLLNVAL))) {
		fprintf( stderr, "%s: read_jobs[%s]: error.\n",
			 argv0, j->name );
		kill_job( j );
	    }	    
	}
    }
}

int
main( int argc, char *argv[] )
{
    char *origbuf;
    ssize_t len;
    char *limit_string;
    struct job *j = NULL;
    struct job *newj;
    int i;

    time( &start_time );

    argv0 = argv[0];

    fprintf( stderr, "%s: t=%d startup.\n", argv0, act_time() );

    limit_string = getenv( "TLIMIT" );
    if (!limit_string) {
	fprintf( stderr, "%s: no TLIMIT in env.\n", argv0 );
	time_limit = 180;
    } else {
	time_limit = atoi( limit_string );
	if (time_limit < 30) {
	    fprintf( stderr, "%s: enforcing minimum TLIMIT of 30.\n", argv0 );
	    time_limit = 30;
	}
    }

    origbuf = read_input( &len );
    notfall_buf = origbuf;
    notfall_buf_len = len;

    best_job = (struct job *)malloc( sizeof( struct job ));
    best_job->name = "<stdin>";
    best_job->running = 0;
    best_job->faulty = 0;
    best_job->readlen = notfall_buf_len;

    signal( SIGPIPE, SIG_IGN );
    signal( SIGALRM, notfall );

    i = time_limit - act_time() - NOTFALL_REST - WARN_TIME;
    fprintf( stderr, "%s: scheduling 1st alarm in %d seconds.\n", argv0, i );
    install_alarm( i );

    newj = spawn( RENDER_PATH, notfall_buf_len + 1 );
    if (newj) {
	if (fill_job( newj, notfall_buf, notfall_buf_len ) == -1) {
	    kill_job( newj );
	} else {
	    newj->next = NULL;
	    j = newj;
	}
    }

    newj = spawn( TREE_PATH, notfall_buf_len + 1 );
    if (newj) {
	if (fill_job( newj, notfall_buf, notfall_buf_len ) == -1) {
	    kill_job( newj );
	} else {
	    warned_job = newj;
	    newj->next = j;
	    j = newj;
	}
    }

    read_jobs( j );

    signal( SIGALRM, SIG_IGN );

    if (best_job && best_job->name)
	fprintf( stderr, "%s: t=%d no jobs left writeout (%s).\n",
		 argv0, act_time(), best_job->name );
    else
	RAUNZ;

    write_output();

    return 0;
}
