open Lexer_types
open List
open Context

exception Hell
exception Contexts_differ
exception Illegal_modifier
exception Illegal_force
exception No_transition

type deco_string =
    Deco_text of string * context
  | Deco_ws of string * context

let modify_context stack tok =
  match tok with
    Close_token tg -> tl stack
  | Open_token tg -> (changed_context (hd stack) tg) :: stack
  | _ -> raise Illegal_modifier

let force o =
  match o with
    Some x -> x
  | None -> raise Illegal_force

let render_with_handler tok h state =
  match state with
    (stack, (deco, h_arg)) ->
      match tok with
	Open_token _ -> (modify_context stack tok, (deco, h_arg))
      | Close_token _ -> (modify_context stack tok, (deco, h_arg))
      | Text_token s ->
	  (stack,
	   match deco with
	     None -> (Some (Deco_text (s, (hd stack))), h_arg)
	   | Some (Deco_text (t, c)) ->
	       if (hd stack) = c then
		 (Some (Deco_text (t ^ s, c)), h_arg)
	       else
		 (Some (Deco_text (s, (hd stack))), h h_arg (Some (Deco_text (t, c))))
	   | Some d ->
	       (Some (Deco_text (s, (hd stack))), h h_arg (Some d)))
      | WS_token s ->
	  let s = if (hd stack).tt_flag then s else " "
	  in (stack,
	      match deco with
		None -> (Some (Deco_ws (s, (hd stack))), h_arg)
	      |	Some (Deco_ws (t, c)) ->
		  if not (contexts_render_ws_same (hd stack) c) then
		    (Some (Deco_ws (s, (hd stack))), h h_arg (Some (Deco_ws (t, c))))
		  else
		    if (hd stack).tt_flag then
		      (Some (Deco_ws (t ^ s, c)), h_arg)
		    else
		      (deco, h_arg)
	      |	Some (Deco_text (t, c)) ->
		  if contexts_render_ws_same (hd stack) c then
		    (Some (Deco_ws (s, c)), h h_arg (Some (Deco_text (t, c))))
		  else
		    (Some (Deco_ws (s, (hd stack))), h h_arg (Some (Deco_text (t, c)))))
      | _ -> raise Hell

let deco_string_string d =
  match d with
    Deco_text (s, c) -> s
  | Deco_ws (s, c) -> s

let deco_string_context d =
  match d with
    Deco_text (s, c) -> c
  | Deco_ws (s, c) -> c

let tag_string tg =
  match tg with
    B_tag -> "B"
  | EM_tag -> "EM"
  | I_tag -> "I"
  | PL_tag -> "PL"
  | S_tag -> "S"
  | TT_tag -> "TT"
  | U_tag -> "U"
  | Size_tag s -> string_of_int s
  | Color_tag Red -> "r"
  | Color_tag Green -> "g"
  | Color_tag Blue -> "b"
  | Color_tag Cyan -> "c"
  | Color_tag Magenta -> "m"
  | Color_tag Yellow -> "y"
  | Color_tag Black -> "k"
  | Color_tag White -> "w"

let print_open tg =
  print_string ("<" ^ (tag_string tg) ^ ">")

let print_close tg =
  print_string ("</" ^ (tag_string tg) ^ ">")

let rec print_closes i stk =
  if i = 0 then
    ()
  else
    let (tg, _) = hd stk
    in print_close tg; print_closes (i - 1) (tl stk)

let tag_length tg =
  2 + String.length (tag_string tg)

let tags_length tgs =
  fold_left (+) 0 (map tag_length tgs)

let close_tags_length tgs =
  (tags_length tgs) + (length tgs)

let transition_length trns =
  tags_length (map (function (tg, _) -> tg) trns)

let rec context_transition c d =
  if (c.b_flag && not d.b_flag) || (c.i_flag && not d.i_flag) || (c.s_flag && not d.s_flag) || (c.tt_flag && not d.tt_flag) || (c.u_level > d.u_level) then
    let n = changed_context c PL_tag
    in (PL_tag, c) :: (context_transition n d)
  else if c.u_level < d.u_level then
    let n = changed_context c U_tag
    in (U_tag, c) :: (context_transition n d)
  else if c.size <> d.size then
    if d.size = None then
      raise No_transition
    else
      let n = { c with size = d.size }
      in (Size_tag (force d.size), c) :: (context_transition n d)
  else if c.color_val <> d.color_val then
    if d.color_val = None then
      raise No_transition
    else
      let n = { c with color_val = d.color_val }
      in (Color_tag (force d.color_val), c) :: (context_transition n d)
  else if (not c.em_flag) && d.em_flag then
    let n = changed_context c EM_tag
    in (EM_tag, c) :: (context_transition n d)
  else if c.em_flag && (not d.em_flag) && d.s_flag then
    let n = changed_context c S_tag
    in (S_tag, c) :: (context_transition n d)
  else if c.em_flag && (not d.em_flag) then
    let t1 = context_transition (changed_context c EM_tag) d
    and t2 = context_transition (changed_context c PL_tag) d
    in if (transition_length t1) < (transition_length t2) then
      (EM_tag, c) :: t1
    else
      (PL_tag, c) :: t2
  else if (not c.b_flag) && d.b_flag then
    let n = changed_context c B_tag
    in (B_tag, c) :: (context_transition n d)
  else if (not c.i_flag) && d.i_flag then
    let n = changed_context c I_tag
    in (I_tag, c) :: (context_transition n d)
  else if (not c.s_flag) && d.s_flag then
    let n = changed_context c S_tag
    in (S_tag, c) :: (context_transition n d)
  else if (not c.tt_flag) && d.tt_flag then
    let n = changed_context c TT_tag
    in (TT_tag, c) :: (context_transition n d)
  else if c <> d then
    raise Contexts_differ
  else
    []

let rec cheapest_transition c stk d max_depth =
  match stk with
    [] -> 
      let t = context_transition c d
      in let cost = transition_length t
      in (map (function x -> let (tg, _) = x in tg) t, rev t, cost, 0)
  | s :: r ->
      let (tg, oc) = s
      in let (pop_t, pop_stk, pop_cost, pop_pops) = cheapest_transition oc r d (max_depth - 1)
      in if max_depth < 0 then
	(pop_t, pop_stk, pop_cost, pop_pops + 1)
      else
	try
	  let t = context_transition c d
	  in let cost = transition_length t
	  in if pop_cost < cost then
	    (pop_t, pop_stk, pop_cost, pop_pops + 1)
	  else
	    (map (function x -> let (tg, _) = x in tg) t, (rev t) @ stk, cost, 0)
	with No_transition ->
	  (pop_t, pop_stk, pop_cost, pop_pops + 1)

let rec remove_nth l n =
  if n = 0 then
    tl l
  else
    (hd l) :: remove_nth (tl l) (n - 1)

let rec position f l =
  match l with
    [] -> 0
  | x :: r -> if f x then 0 else 1 + position f r;;

let rec build_stack stk c tgs =
  match tgs with
    [] -> stk
  | tg :: r -> build_stack ((tg, c) :: stk) (changed_context c tg) r

let rec contexts ds c =
  match ds with
    [] -> if c = root_context then [c] else [c; root_context]
  | d :: r ->
      let dc = deco_string_context d
      in if dc = c then contexts r c else c :: (contexts r dc)

let map_with_remove f l =
  let rec work rev_head l =
    match l with
      [] -> []
    | x :: r ->
	(f x (rev_append rev_head r)) :: (work (x :: rev_head) r)
  in work [] l

let rec unrender_handler arg deco_opt =
  let (stk, ctx) = arg
  in match deco_opt with
    None -> iter (function (tg, _) -> print_close tg) stk; ([], root_context)
  | Some d ->
      let (t, new_stk, _, pops) = cheapest_transition ctx stk (deco_string_context d) 15
      in print_closes pops stk;
      iter print_open t;
      print_string (deco_string_string d);
      (new_stk, deco_string_context d)

let rec parse_with_unrender lexbuf state =
  let tok = Lexer.scan lexbuf
  in match tok with
    EOF -> (match state with (_, (deco, h_arg)) -> unrender_handler (unrender_handler h_arg deco) None)
  | _ -> parse_with_unrender lexbuf (render_with_handler tok unrender_handler state)

let rec read_on s p m =
  let num = Unix.read (Unix.descr_of_in_channel stdin) s p (m - p)
  in if num = 0 then
    String.sub s 0 p
  else
    read_on s (p + num) m

let read_whole_input () =
  read_on (String.create (5 * 1024 * 1024)) 0 (5 * 1024 * 1024)

let main () =
  let lexbuf = Lexing.from_string (read_whole_input ())
  in parse_with_unrender lexbuf ([root_context], (None, ([], root_context)));;

main ();;
exit 0;;
