(* -*- fundamental -*- *)

open List
open Lexer_types
open Context
open Parser
open Debug
open Util
open NonDetMonad 
open NonDetUtils
open Sys
open Printf

exception OutOfTime

let successful_matches = ref 0

(****************************************************************************)

let (@+) = append

let (@-) = rev_append

let tagShadows a b = match (a,b) with
  | (Color_tag _, Color_tag _)
  | (Size_tag _,  Size_tag _)
  | (EM_tag, PL_tag)
  | (EM_tag, S_tag)
  | (I_tag, PL_tag)
  | (B_tag, PL_tag)
  | (U_tag, PL_tag)
  | (S_tag, PL_tag)
  | (TT_tag, PL_tag) -> true
  | _ -> false

let rec tagsShadowOptimized = function
  | [] -> []
  | x::xs ->
      if exists (tagShadows x) xs then
	tagsShadowOptimized xs
      else
	x::(tagsShadowOptimized xs)

(****************************************************************************)

let rec optimize_insert_combnodes bag flag = function
  | [] -> if flag then unitM (concat (rev bag)) else failM
  | (Node(_,[],succs))::xs -> optimize_insert_combnodes (succs::bag) true xs
  | x::xs -> optimize_insert_combnodes ([x]::bag) flag xs


let optimize_whitespaces1 current_context successors =
  let rec loopM bag flag = function
    | (Node(_,ts1,succs1))::(WhiteSpace _ as ws)::xs
      when (let context_in_t1 = fold_left changed_context current_context ts1 in
	      context_in_t1.u_level = current_context.u_level &&
	      context_in_t1.size = current_context.size &&
	      (if context_in_t1.u_level > 0 then 
		 context_in_t1.color_val = current_context.color_val
	       else 
	         true) &&
	       context_in_t1.tt_flag = current_context.tt_flag) ->
	loopM ((Node(true,ts1,succs1@[ws]))::bag) true xs
    | x::xs ->
	loopM (x::bag) flag xs
    | [] ->
	if flag then unitM (rev bag) else failM
  in loopM [] false successors


let optimize_whitespaces2 current_context successors =
  let rec loopM bag flag = function
    | (WhiteSpace _ as ws)::(Node(_,ts1,succs1))::xs
      when (let context_in_t1 = fold_left changed_context current_context ts1 in
	      context_in_t1.u_level = current_context.u_level &&
	      context_in_t1.size = current_context.size &&
	      (if context_in_t1.u_level > 0 then 
		 context_in_t1.color_val = current_context.color_val
	       else 
	         true) &&
	       context_in_t1.tt_flag = current_context.tt_flag) ->
	loopM ((Node(true,ts1,ws::succs1))::bag) true xs
    | x::xs ->
	loopM (x::bag) flag xs
    | [] ->
	if flag then unitM (rev bag) else failM
  in loopM [] false successors


let optimize_whitespace_text successors =
  let rec loopM bag flag = function
    | (WhiteSpace a)::(Text b)::xs 
    | (Text a)::(WhiteSpace b)::xs ->
	loopM bag true ((Text (a^b))::xs)
    | x::xs ->
	loopM (x::bag) flag xs
    | [] ->
	if flag then unitM (rev bag) else failM
  in loopM [] false successors


let optimize_join_neighboring_nodes_rigid current_context successors =
  let rec loopM bag flag = function
    | (Node(_,ts1,succs1))::(Node(_,ts2,succs2))::xs
      when (fold_left changed_context current_context ts1 =
	    fold_left changed_context current_context ts2) ->

	let ts_shorter = if length ts1 > length ts2 then ts2 else ts1 in
	  loopM ((Node(true,ts_shorter,succs1@succs2))::bag) true xs
    | x::xs ->
	loopM (x::bag) flag xs
    | [] ->
	if flag then unitM (rev bag) else failM
  in loopM [] false successors
        

let optimize_eliminate_useless_nodes successors =
  let rec loopM bag flag = function
    | (Node(_,_,[]))::xs ->
	loopM bag true xs
    | x::xs ->
	loopM (x::bag) flag xs
    | [] ->
	if flag then unitM (rev bag) else failM
  in loopM [] false successors

(****************************************************************************)

let optimization_rules =
  [
   "[P2] join two nodes: node(t1,[node(t2,succs)]) ==> node(t1@t2,succs)",
   (fun (current_context,successors,t1) ->
	posAssertM (t1<>[]) >> 
      match successors with
	| [Node(_,t2,succs)] -> unitM (Node(true,t1@t2,succs))
	| _ -> failM
   );

   "whitespace optimization 1",
   (fun (current_context,successors,tags) ->
      optimize_whitespaces1 current_context successors >>= fun successors' ->
	unitM (Node(true,tags,successors'))
   );

   "whitespace optimization 2",
   (fun (current_context,successors,tags) ->
      optimize_whitespaces2 current_context successors >>= fun successors' ->
	unitM (Node(true,tags,successors'))
   );

   "combine whitespace+text (or text+whitespace)",
   (fun (_,successors,tags) ->
      optimize_whitespace_text successors >>= fun successors' ->
	unitM (Node(true,tags,successors'))
   );

   "insert combnodes: node(t1,[...,node([],succs),...]) ==> node(t1,[... @ succs @ ...])",
   (fun (current_context,successors,tags) ->
      optimize_insert_combnodes [] false successors >>= fun successors' ->
	unitM (Node(true,tags,successors'))
   );

   "eliminate useless nodes",
   (fun (_,successors,tags) ->
      optimize_eliminate_useless_nodes successors >>= fun successors' ->
	unitM (Node(true,tags,successors'))
   );

   "join neighboring nodes (rigid)",
   (fun (current_context,successors,tags) ->
      optimize_join_neighboring_nodes_rigid current_context successors >>= fun succs' ->
	unitM (Node(true,tags,succs'))
   );

   "join neighboring nodes (more liberally)",
   (fun (current_context,successors,tags) ->
      select4M successors >>= function
	| (pre', (Node(_,ts1,succs1)), (Node(_,ts2,succs2)), post) ->
	     select3M ts1 >>= fun (pre1', t1, post1) ->
	       posAssertM 
		  (fold_left changed_context current_context (pre1' @- [t1]) =
		   fold_left changed_context current_context (t1::(rev pre1'))) >>
	     select3M ts2 >>= fun (pre2', t2, post2) ->
	       posAssertM 
		  (fold_left changed_context current_context (pre2' @- [t2]) =
		   fold_left changed_context current_context (t2::(rev pre2'))) >>
		 posAssertM (t1=t2) >>

		   unitM (Node(true,
			       tags,pre' @- [Node(true,
						  [t1],[Node(true,pre1' @- post1,succs1);
							Node(true,pre2' @- post2,succs2)])]
					 @+ post))
	| _ -> failM      
   );

   "nesting of colors and sizes (left)",
   (fun (_,successors,tags) ->
      select4M successors >>= function
	| (pre',(Node(_,[t1],succs1)),(Node(_,[t2],_) as node2),post) ->
	    (match (t1,t2) with
	       | (Color_tag _, Color_tag _) 
	       | (Size_tag _,  Size_tag _) -> unitM ()
	       | _ -> failM) >>
	    unitM (Node(true,tags,pre' @- [Node(true,[t1],succs1@[node2])] @+ post))
	| _ -> failM
   );

   "overlap optimization",
   (fun (current_context,successors,tags) ->
      select4M successors >>= function
	| (pre',(Node(_,ts1,succs1)),(Node(_,ts2,(Node(_,ts3,succs3))::succs2)),post) ->
	    posAssertM (length ts1 > 1) >>
	    posAssertM (fold_left changed_context current_context (ts1 @ ts2) =
			fold_left changed_context current_context (ts2 @ ts3)) >>
	    unitM (Node(true,
			tags,
			pre' @- [Node(true,ts1,succs1 @ [Node(true,ts2,succs3)]); 
				 Node(true,ts2,succs2)] 
			     @+ post))
	| _ -> failM
   );


   "eliminate locally-shadowed tags",
   (fun (_,successors,tags) ->
      let tags' = tagsShadowOptimized tags in
	posAssertM (tags <> tags') >>
	  unitM (Node(true,tags',successors))
   );

   "eliminate useless tags (simple)",
   (fun (current_context,successors,tags) ->
	failM >>
      select3M successors >>= function
	| (pre', (Node(_,ts1,succs1)), post) ->
	    let context_in_t1  = fold_left changed_context current_context ts1 in
	      posAssertM (current_context = context_in_t1) >>
		unitM (Node(true,tags,pre' @- succs1 @+ post))
	| _ -> failM
   );

   "eliminate <EM>",
   (fun (current_context,successors,tags) ->
      select3M tags >>= function
	| (pre',EM_tag,post) ->
	    (select3M post >>= function
	       | (pre2',EM_tag,post2) ->
		   posAssertM (not (exists (tagShadows EM_tag) pre2')) >>
		     unitM (Node(true,(pre2' @+ pre') @- post2,successors))
	       | _ -> failM)
	| _ -> failM
   );

   "[P2] PL shortcut",
   (fun (current_context,successors,tags) ->
      select5M successors >>= function
	| (pre',(Node(_,ts1,succs1)),node2,(Node(_,ts3,succs3)),post) ->
	    let context_in_t1 = fold_left changed_context current_context ts1
	    and context_in_t3 = fold_left changed_context current_context ts3 in
	      posAssertM (context_in_t1 = context_in_t3) >>
	      posAssertM (changed_context context_in_t1 PL_tag = current_context) >>
		let shorter_tags = 
			if List.length ts1 > List.length ts3 then
			  ts3
			else 
			  ts1 in
	  	posAssertM (List.length shorter_tags > 1) >>
		  unitM (Node(true,
			      tags,
			      pre' @- [Node(true,
					    shorter_tags,
					    succs1 @ [Node(true,[PL_tag],[node2])] @ succs3)] 
				   @+ post))
	| _ -> failM
   );

   "[P2] propagate PL-up (left)",
   (fun (current_context,successors,tags) ->
      select3M successors >>= function
	| (pre', (Node(_,ts1,(Node(_,ts2,_) as node2)::succs1)), post) ->
	    posAssertM (mem PL_tag ts2) >>= fun _ ->	(* for better termination *)
	    let context_in_t1  = fold_left changed_context current_context ts1 in
	    let context_in_t2  = fold_left changed_context context_in_t1 ts2 in
	    let context_in_t2' = fold_left changed_context current_context ts2 in
	      posAssertM (context_in_t2 = context_in_t2') >>
	 	unitM (Node(true,tags,pre' @- [node2; Node(true,ts1,succs1)] @+ post))
	| _ -> failM
   );

   "[P2] propagate PL-up (right)",
   (fun (current_context,successors,tags) ->
      select3M successors >>= function
	| (pre', (Node(_,ts1,(_::_ as succs1))), post) ->
	    let (last1,succs1) = list_butlast succs1 in
	      (match last1 with
		 | Node(_,ts2,_) as node2 ->
		     posAssertM (mem PL_tag ts2) >>	(* for better termination *)
		       let context_in_t1  = fold_left changed_context current_context ts1 in
		       let context_in_t2  = fold_left changed_context context_in_t1 ts2 in
		       let context_in_t2' = fold_left changed_context current_context ts2 in
		         posAssertM (context_in_t2 = context_in_t2') >>
		 	   unitM (Node(true,tags,pre' @- [Node(true,ts1,succs1); node2] @+ post))
	 	 | _ -> failM)
	| _ -> failM
   );
  ]


let optimizeNodeM' current_context successors tags =
  memberM optimization_rules >>= fun (rule_description,fM) ->
    fM (current_context,successors,tags) >>= fun x ->
      let _ = debugOutputString (Printf.sprintf " used rule: %s." rule_description) in  
	unitM x

let isUnoptimizableNode = function
  | Node(false,_,_) 
  | Text _ 
  | WhiteSpace _ -> true
  | Node(true,_,_) -> false

let rec optimizeNodeM current_context successors tags =
  (   (fun _ ->
	 onceM (optimizeNodeM' current_context successors) tags >>= fun new_node ->
	   incr successful_matches;
	     unitM new_node		(* is marked as optimizable anyway *)
      )
  ||| (fun _ ->				
	 unitM (Node(not (for_all isUnoptimizableNode successors),tags,successors))
      )
  )

let rec matcherM current_context = function
  | Text s -> 
      unitM (Text s)
  | WhiteSpace s  -> 
      unitM (WhiteSpace s)
  | Node(false,_,_) as node -> 
      unitM node
  | Node(_,tags,successors)  -> 
      let new_context = fold_left changed_context current_context tags in
	mapM (matcherM new_context) successors >>= fun successors' ->
	  optimizeNodeM new_context successors' tags


(****************************************************************************)

let current_best_solution = ref None

let optionToValue = function 
  | Some a -> a
  | None   -> failwith "optionToValue"

let rec matcher' tree =
  successful_matches := 0;
  let tree' = optionToValue (NonDetMonad.runM (matcherM root_context) tree) in
  debugOutputString (sprintf "# of matches: %d.\n" !successful_matches);
  if !successful_matches > 0 then
  begin
    current_best_solution := Some tree';
    matcher' tree'
  end
  else 
    tree

let matcher tree =
  ignore (signal sigusr1 (Signal_handle (fun _ -> raise OutOfTime)));
  current_best_solution := Some tree;
  try
    matcher' tree
  with OutOfTime ->
    optionToValue !current_best_solution
